import numpy as np
from numpy.linalg import lstsq

from amodem import dsp
from amodem import sampling

import itertools
import random

_constellation = [1, 1j, -1, -1j]


class Equalizer(object):

    def __init__(self, config):
        self.carriers = config.carriers
        self.omegas = 2 * np.pi * np.array(config.frequencies) / config.Fs
        self.Nfreq = config.Nfreq
        self.Nsym = config.Nsym

    def train_symbols(self, length, seed=0):
        r = random.Random(seed)
        choose = lambda: [r.choice(_constellation) for j in range(self.Nfreq)]
        return np.array([choose() for _ in range(length)])

    def modulator(self, symbols):
        gain = 1.0 / len(self.carriers)
        result = []
        for s in symbols:
            result.append(np.dot(s, self.carriers))
        result = np.concatenate(result).real * gain
        assert np.max(np.abs(result)) <= 1
        return result

    def demodulator(self, signal, size):
        signal = itertools.chain(signal, itertools.repeat(0))
        symbols = dsp.Demux(sampler=sampling.Sampler(signal),
                            omegas=self.omegas, Nsym=self.Nsym)
        return np.array(list(itertools.islice(symbols, size)))


def train(signal, expected, order, lookahead=0):
    signal = [np.zeros(order-1), signal, np.zeros(lookahead)]
    signal = np.concatenate(signal)
    length = len(expected)

    A = []
    b = []

    for i in range(length - order):
        offset = order + i
        row = signal[offset-order:offset+lookahead]
        A.append(np.array(row, ndmin=2))
        b.append(expected[i])

    A = np.concatenate(A, axis=0)
    b = np.array(b)
    h = lstsq(A, b)[0]
    h = h[::-1].real
    return h


prefix = [1]*400 + [0]*50
equalizer_length = 500
silence_length = 100



import itertools
import numpy as np

import logging
log = logging.getLogger(__name__)

scaling = 32000.0  # out of 2**15


def load(fileobj):
    return loads(fileobj.read())


def loads(data):
    x = np.frombuffer(data, dtype='int16')
    x = x / scaling
    return x


def dumps(sym):
    sym = sym.real * scaling
    return sym.astype('int16').tostring()


def iterate(data, size, func=None, truncate=True, index=False):
    offset = 0
    data = iter(data)

    done = False
    while not done:
        buf = list(itertools.islice(data, size))
        if len(buf) < size:
            if truncate or not buf:
                return
            done = True

        result = func(buf) if func else np.array(buf)
        yield (offset, result) if index else result
        offset += size


def split(iterable, n):
    def _gen(it, index):
        for item in it:
            yield item[index]

    iterables = itertools.tee(iterable, n)
    return [_gen(it, index) for index, it in enumerate(iterables)]


def icapture(iterable, result):
    for i in iter(iterable):
        result.append(i)
        yield i


def take(iterable, n):
    return np.array(list(itertools.islice(iterable, n)))


# "Python 3" zip re-implementation for Python 2
def izip(iterables):
    iterables = [iter(iterable) for iterable in iterables]
    while True:
        yield tuple([next(iterable) for iterable in iterables])


class Dummy(object):

    def __getattr__(self, name):
        return self

    def __call__(self, *args, **kwargs):
        return self


class AttributeHolder(object):

    def __init__(self, d):
        self.__dict__.update(d)

    def __repr__(self):
        items = sorted(self.__dict__.items())
        args = ', '.join('{0}={1}'.format(k, v) for k, v in items)
        return '{0}({1})'.format(self.__class__.__name__, args)



import time


class Reader(object):

    wait = 0.2
    timeout = 2.0
    bufsize = 4096

    def __init__(self, fd, data_type=None, eof=False):
        self.fd = fd
        self.data_type = data_type if (data_type is not None) else lambda x: x
        self.eof = eof
        self.total = 0

    def __iter__(self):
        return self

    def next(self):
        block = bytearray()
        if self.eof:
            data = self.fd.read(self.bufsize)
            if data:
                self.total += len(data)
                block.extend(data)
                return block
            else:
                raise StopIteration()

        finish_time = time.time() + self.timeout
        while time.time() <= finish_time:
            left = self.bufsize - len(block)
            data = self.fd.read(left)
            if data:
                self.total += len(data)
                block.extend(data)

            if len(block) == self.bufsize:
                return self.data_type(block)

            time.sleep(self.wait)

        raise IOError('timeout')

    __next__ = next



import numpy as np
import itertools
import logging

log = logging.getLogger(__name__)

from . import common
from . import dsp
from . import sampling

ALLOWED_EXCEPTIONS = (IOError, KeyboardInterrupt)


def send(config, dst):
    calibration_symbols = int(1.0 * config.Fs)
    t = np.arange(0, calibration_symbols) * config.Ts
    signals = [np.sin(2 * np.pi * f * t) for f in config.frequencies]
    signals = [common.dumps(s) for s in signals]

    try:
        for signal in itertools.cycle(signals):
            dst.write(signal)
    except ALLOWED_EXCEPTIONS:
        pass


def frame_iter(config, src, frame_length):
    frame_size = frame_length * config.Nsym * config.sample_size
    omegas = 2 * np.pi * config.frequencies / config.Fs

    while True:
        data = src.read(frame_size)
        if len(data) < frame_size:
            return
        data = common.loads(data)
        frame = data - np.mean(data)

        sampler = sampling.Sampler(frame)
        symbols = dsp.Demux(sampler, omegas, config.Nsym)

        symbols = np.array(list(symbols))
        coeffs = np.mean(np.abs(symbols) ** 2, axis=0) ** 0.5

        peak = np.max(np.abs(frame))
        total = np.sqrt(np.dot(frame, frame) / (0.5 * len(frame)))
        yield coeffs, peak, total


def detector(config, src, frame_length=200):

    states = [True]
    errors = ['weak', 'strong', 'noisy']
    try:
        for coeffs, peak, total in frame_iter(config, src, frame_length):
            max_index = np.argmax(coeffs)
            freq = config.frequencies[max_index]
            rms = abs(coeffs[max_index])
            coherency = rms / total
            flags = [rms > 0.1, peak < 1.0, coherency > 0.99]

            states.append(all(flags))
            states = states[-2:]

            message = 'good signal'
            error = not any(states)
            if error:
                message = 'too {0} signal'.format(errors[flags.index(False)])

            yield common.AttributeHolder(dict(
                freq=freq, rms=rms, peak=peak, coherency=coherency,
                total=total, error=error, message=message
            ))
    except ALLOWED_EXCEPTIONS:
        pass


def recv(config, src, verbose=False):
    fmt = '{0.freq:6.0f} Hz: {0.message:s}'
    if verbose:
        fields = ['peak', 'total', 'rms', 'coherency']
        fmt += ''.join(', {0}={{0.{0}:.4f}}'.format(f) for f in fields)

    for result in detector(config=config, src=src):
        msg = fmt.format(result)
        if not result.error:
            log.info(msg)
        else:
            log.error(msg)



import numpy as np
from numpy import linalg

from . import common


class FIR(object):
    def __init__(self, h):
        self.h = np.array(h)
        self.x_state = [0] * len(self.h)

    def __call__(self, x):
        x_ = self.x_state
        h = self.h
        for v in x:
            x_ = [v] + x_[:-1]
            yield np.dot(x_, h)
        self.x_state = x_


class IIR(object):
    def __init__(self, b, a):
        self.b = np.array(b) / a[0]
        self.a = np.array(a[1:]) / a[0]
        self.x_state = [0] * len(self.b)
        self.y_state = [0] * (len(self.a) + 1)

    def __call__(self, x):
        x_, y_ = self.x_state, self.y_state
        for v in x:
            x_ = [v] + x_[:-1]
            y_ = y_[:-1]
            num = np.dot(x_, self.b)
            den = np.dot(y_, self.a)
            y = num - den
            y_ = [y] + y_
            yield y
        self.x_state, self.y_state = x_, y_


def lfilter(b, a, x):
    f = IIR(b=b, a=a)
    y = list(f(x))
    return np.array(y)


class Demux(object):
    def __init__(self, sampler, omegas, Nsym):
        self.Nsym = Nsym
        self.filters = [exp_iwt(-w, Nsym) / (0.5*self.Nsym) for w in omegas]
        self.filters = np.array(self.filters)
        self.sampler = sampler

    def __iter__(self):
        return self

    def next(self):
        frame = self.sampler.take(size=self.Nsym)
        if len(frame) == self.Nsym:
            return np.dot(self.filters, frame)
        else:
            raise StopIteration

    __next__ = next


def exp_iwt(omega, n):
    return np.exp(1j * omega * np.arange(n))


def norm(x):
    return np.sqrt(np.dot(x.conj(), x).real)


def coherence(x, omega):
    n = len(x)
    Hc = exp_iwt(-omega, n) / np.sqrt(0.5*n)
    norm_x = norm(x)
    if norm_x:
        return np.dot(Hc, x) / norm_x
    else:
        return 0.0


def linear_regression(x, y):
    ''' Find (a,b) such that y = a*x + b. '''
    x = np.array(x)
    y = np.array(y)
    ones = np.ones(len(x))
    M = np.array([x, ones]).T
    a, b = linalg.lstsq(M, y)[0]
    return a, b


class MODEM(object):

    buf_size = 16

    def __init__(self, symbols):
        self.encode_map = {}
        symbols = np.array(list(symbols))
        bits_per_symbol = np.log2(len(symbols))
        bits_per_symbol = np.round(bits_per_symbol)
        N = (2 ** bits_per_symbol)
        assert N == len(symbols)
        bits_per_symbol = int(bits_per_symbol)

        for i, v in enumerate(symbols):
            bits = [int(i & (1 << j) != 0) for j in range(bits_per_symbol)]
            self.encode_map[tuple(bits)] = v

        self.symbols = symbols
        self.bits_per_symbol = bits_per_symbol

        bits_map = dict(item[::-1] for item in self.encode_map.items())
        self.decode_list = [(s, bits_map[s]) for s in self.symbols]

    def encode(self, bits):
        for bits_tuple in common.iterate(bits, self.bits_per_symbol, tuple):
            yield self.encode_map[bits_tuple]

    def decode(self, symbols, error_handler=None):
        ''' Maximum-likelihood decoding, using naive nearest-neighbour. '''
        symbols_vec = self.symbols
        _dec = self.decode_list
        for syms in common.iterate(symbols, self.buf_size, truncate=False):
            for received in syms:
                error = np.abs(symbols_vec - received)
                index = np.argmin(error)
                decoded, bits = _dec[index]
                if error_handler:
                    error_handler(received=received, decoded=decoded)
                yield bits



import numpy as np


class Configuration(object):
    Fs = 32000.0  # sampling frequency [Hz]
    Tsym = 0.001  # symbol duration [seconds]
    Nfreq = 8     # number of frequencies used
    Npoints = 64
    F0 = 1e3

    # audio config
    bits_per_sample = 16
    sample_size = bits_per_sample // 8
    samples_per_buffer = 4096

    # sender config
    silence_start = 1.0
    silence_stop = 1.0

    # receiver config
    skip_start = 0.1

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

        self.Ts = 1.0 / self.Fs
        self.Fsym = 1 / self.Tsym
        self.frequencies = self.F0 + np.arange(self.Nfreq) * self.Fsym
        self.carrier_index = 0
        self.Fc = self.frequencies[self.carrier_index]

        self.Nsym = int(self.Tsym / self.Ts)
        self.baud = int(1.0 / self.Tsym)

        bits_per_symbol = np.log2(self.Npoints)
        assert int(bits_per_symbol) == bits_per_symbol
        self.bits_per_baud = bits_per_symbol * self.Nfreq
        self.modem_bps = self.baud * self.bits_per_baud
        self.carriers = np.array([
            np.exp(2j * np.pi * f * np.arange(0, self.Nsym) * self.Ts)
            for f in self.frequencies
        ])

        # QAM constellation
        Nx = 2 ** int(np.ceil(bits_per_symbol / 2))
        Ny = self.Npoints // Nx
        symbols = [complex(x, y) for x in range(Nx) for y in range(Ny)]
        symbols = np.array(symbols)
        symbols = symbols - symbols[-1]/2
        self.symbols = symbols / np.max(np.abs(symbols))

# MODEM configurations for various bitrates [kbps]
bitrates = {
    1: Configuration(F0=8e3, Npoints=2, Nfreq=1),
    2: Configuration(F0=8e3, Npoints=4, Nfreq=1),
    4: Configuration(F0=8e3, Npoints=16, Nfreq=1),
    8: Configuration(F0=8e3, Npoints=16, Nfreq=2),
    16: Configuration(F0=6e3, Npoints=16, Nfreq=4),
    32: Configuration(F0=3e3, Npoints=16, Nfreq=8),
    48: Configuration(F0=3e3, Npoints=64, Nfreq=8),
    54: Configuration(F0=3e3, Npoints=64, Nfreq=9),
    60: Configuration(F0=2e3, Npoints=64, Nfreq=10),
}

fastest = lambda: bitrates[max(bitrates)]
slowest = lambda: bitrates[min(bitrates)]



import ctypes
import logging

log = logging.getLogger(__name__)


class Interface(object):
    def __init__(self, name, config, debug=False):
        self.debug = bool(debug)
        self.lib = ctypes.CDLL(name)
        self.config = config
        self.streams = []
        assert self._error_string(0) == 'Success'

    def _error_string(self, code):
        return self.call('GetErrorText', code, restype=ctypes.c_char_p)

    def call(self, name, *args, **kwargs):
        func_name = 'Pa_{0}'.format(name)
        if self.debug:
            log.debug('API: %s%s', name, args)
        func = getattr(self.lib, func_name)
        func.restype = kwargs.get('restype', self._error_check)
        return func(*args)

    def _error_check(self, res):
        if res != 0:
            raise Exception(res, self._error_string(res))

    def __enter__(self):
        self.call('Initialize')
        return self

    def __exit__(self, *args):
        for s in self.streams:
            s.close()
        self.call('Terminate')

    def recorder(self):
        return Stream(lib=self, config=self.config, read=True)

    def player(self):
        return Stream(lib=self, config=self.config, write=True)


class Stream(object):

    class Parameters(ctypes.Structure):
        _fields_ = [
            ('device', ctypes.c_int),
            ('channelCount', ctypes.c_int),
            ('sampleFormat', ctypes.c_ulong),
            ('suggestedLatency', ctypes.c_double),
            ('hostApiSpecificStreamInfo', ctypes.POINTER(None))
        ]

    def __init__(self, lib, config, read=False, write=False):
        self.lib = lib
        self.stream = ctypes.POINTER(ctypes.c_void_p)()
        self.user_data = ctypes.c_void_p(None)
        self.stream_callback = ctypes.c_void_p(None)
        self.bytes_per_sample = config.sample_size
        assert config.bits_per_sample == 16  # just to make sure :)

        read = bool(read)
        write = bool(write)
        assert read != write

        direction = 'Input' if read else 'Output'
        api_name = 'GetDefault{0}Device'.format(direction)
        index = lib.call(api_name, restype=ctypes.c_int)
        self.params = Stream.Parameters(
            device=index,               # choose default device
            channelCount=1,             # mono audio
            sampleFormat=0x00000008,    # 16-bit samples (paInt16)
            suggestedLatency=0.1,       # 100ms should be good enough
            hostApiSpecificStreamInfo=None)

        self.lib.call(
            'OpenStream',
            ctypes.byref(self.stream),
            ctypes.byref(self.params) if read else None,
            ctypes.byref(self.params) if write else None,
            ctypes.c_double(config.Fs),
            ctypes.c_ulong(config.samples_per_buffer),
            ctypes.c_ulong(0),  # no flags (paNoFlag)
            self.stream_callback,
            self.user_data)

        self.lib.streams.append(self)
        self.lib.call('StartStream', self.stream)

    def close(self):
        if self.stream:
            self.lib.call('StopStream', self.stream)
            self.lib.call('CloseStream', self.stream)
            self.stream = None

    def read(self, size):
        assert size % self.bytes_per_sample == 0
        buf = ctypes.create_string_buffer(size)
        frames = ctypes.c_ulong(size // self.bytes_per_sample)
        self.lib.call('ReadStream', self.stream, buf, frames)
        return buf.raw

    def write(self, data):
        data = bytes(data)
        assert len(data) % self.bytes_per_sample == 0
        buf = ctypes.c_char_p(data)
        frames = ctypes.c_ulong(len(data) // self.bytes_per_sample)
        self.lib.call('WriteStream', self.stream, buf, frames)



import numpy as np
import logging
import itertools

log = logging.getLogger(__name__)

from . import common
from . import stream
from . import framing
from . import equalizer
from . import dsp


class Sender(object):
    def __init__(self, fd, config):
        self.offset = 0
        self.fd = fd
        self.modem = dsp.MODEM(config.symbols)
        self.carriers = config.carriers / config.Nfreq
        self.pilot = config.carriers[config.carrier_index]
        self.silence = np.zeros(equalizer.silence_length * config.Nsym)
        self.iters_per_report = config.baud  # report once per second
        self.padding = [0] * config.bits_per_baud
        self.equalizer = equalizer.Equalizer(config)

    def write(self, sym):
        sym = np.array(sym)
        data = common.dumps(sym)
        self.fd.write(data)
        self.offset += len(sym)

    def start(self):
        for value in equalizer.prefix:
            self.write(self.pilot * value)

        symbols = self.equalizer.train_symbols(equalizer.equalizer_length)
        signal = self.equalizer.modulator(symbols)
        self.write(self.silence)
        self.write(signal)
        self.write(self.silence)

    def modulate(self, bits):
        bits = itertools.chain(bits, self.padding)
        Nfreq = len(self.carriers)
        symbols_iter = common.iterate(self.modem.encode(bits), size=Nfreq)
        for i, symbols in enumerate(symbols_iter, 1):
            self.write(np.dot(symbols, self.carriers))
            if i % self.iters_per_report == 0:
                total_bits = i * Nfreq * self.modem.bits_per_symbol
                log.debug('Sent %10.3f kB', total_bits / 8e3)


def main(config, src, dst):
    sender = Sender(dst, config=config)
    Fs = config.Fs

    # pre-padding audio with silence
    sender.write(np.zeros(int(Fs * config.silence_start)))

    sender.start()

    training_duration = sender.offset
    log.info('Sending %.3f seconds of training audio', training_duration / Fs)

    reader = stream.Reader(src, eof=True)
    data = itertools.chain.from_iterable(reader)
    bits = framing.encode(data)
    log.info('Starting modulation')
    sender.modulate(bits=bits)

    data_duration = sender.offset - training_duration
    log.info('Sent %.3f kB @ %.3f seconds',
             reader.total / 1e3, data_duration / Fs)

    # post-padding audio with silence
    sender.write(np.zeros(int(Fs * config.silence_stop)))
    return True



#!/usr/bin/env python
from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):

    def finalize_options(self):
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import sys
        import pytest
        sys.exit(pytest.main(['tests']))

setup(
    name="amodem",
    version="1.6",
    description="Audio Modem Communication Library",
    author="Roman Zeyde",
    author_email="roman.zeyde@gmail.com",
    license="MIT",
    url="http://github.com/romanz/amodem",
    packages=['amodem'],
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    install_requires=['numpy'],
    platforms=['POSIX'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Networking",
        "Topic :: Communications",
    ],
    scripts=['amodem-cli'],
)



#!/usr/bin/env python
import numpy as np
import itertools

from amodem import common


class Interpolator(object):

    def __init__(self, resolution=1024, width=128):

        self.width = width
        self.resolution = resolution

        N = resolution * width
        u = np.arange(-N, N, dtype=float)
        window = (1 + np.cos(0.5 * np.pi * u / N)) / 2.0  # (Hann window)

        h = np.sinc(u / resolution) * window
        self.filt = []
        for index in range(resolution):  # split into multiphase filters
            filt = h[index::resolution]
            filt = filt[::-1]  # flip (due to convolution)
            self.filt.append(filt)

        lengths = [len(f) for f in self.filt]
        self.coeff_len = 2 * width

        assert set(lengths) == set([self.coeff_len])  # verify same lengths
        assert len(self.filt) == resolution


class Sampler(object):
    def __init__(self, src, interp=None, freq=1.0):
        self.freq = freq
        self.equalizer = lambda x: x  # LTI equalization filter
        if interp is not None:
            self.interp = interp
            self.resolution = self.interp.resolution
            self.filt = self.interp.filt
            self.width = self.interp.width

            # polyphase filters are centered at (width + 1) index
            padding = [0.0] * self.interp.width
            # pad with zeroes to "simulate" regular sampling
            self.src = itertools.chain(padding, src)
            self.offset = self.interp.width + 1
            # samples' buffer to be used by interpolation
            self.buff = np.zeros(self.interp.coeff_len)
            self.index = 0
            self.take = self._take
        else:
            # skip interpolation (for testing)
            src = iter(src)
            self.take = lambda size: common.take(src, size)

    def _take(self, size):
        frame = np.zeros(size)
        count = 0
        try:
            for frame_index in range(size):
                offset = self.offset
                # offset = k + (j / self.resolution)
                k = int(offset)  # integer part
                j = int((offset - k) * self.resolution)  # fractional part
                coeffs = self.filt[j]  # choose correct filter phase
                end = k + self.width
                # process input until all buffer is full with samples
                while self.index < end:
                    self.buff[:-1] = self.buff[1:]
                    self.buff[-1] = next(self.src)  # throws StopIteration
                    self.index += 1

                self.offset += self.freq
                # apply interpolation filter
                frame[frame_index] = np.dot(coeffs, self.buff)
                count = frame_index + 1
        except StopIteration:
            pass

        return self.equalizer(frame[:count])


def resample(src, dst, df=0.0):
    x = common.load(src)
    sampler = Sampler(x, Interpolator())
    sampler.freq += df
    y = sampler.take(len(x))
    dst.write(common.dumps(y))



from . import common

import functools
import itertools
import binascii
import struct
import logging
log = logging.getLogger(__name__)

_checksum_func = lambda x: binascii.crc32(bytes(x)) & 0xFFFFFFFF
# (so the result will be unsigned on Python 2/3)


class Checksum(object):
    fmt = '>L'  # unsigned longs (32-bit)
    size = struct.calcsize(fmt)

    def encode(self, payload):
        checksum = _checksum_func(payload)
        return struct.pack(self.fmt, checksum) + payload

    def decode(self, data):
        received, = struct.unpack(self.fmt, bytes(data[:self.size]))
        payload = data[self.size:]
        expected = _checksum_func(payload)
        if received != expected:
            log.warning('Invalid checksum: %04x != %04x', received, expected)
            raise ValueError('Invalid checksum')
        return payload


class Framer(object):
    block_size = 250
    prefix_fmt = '>B'
    prefix_len = struct.calcsize(prefix_fmt)
    checksum = Checksum()

    EOF = b''

    def _pack(self, block):
        frame = self.checksum.encode(block)
        return bytearray(struct.pack(self.prefix_fmt, len(frame)) + frame)

    def encode(self, data):
        for block in common.iterate(data=data, size=self.block_size,
                                    func=bytearray, truncate=False):
            yield self._pack(block=block)
        yield self._pack(block=self.EOF)

    def decode(self, data):
        data = iter(data)
        while True:
            length, = _take_fmt(data, self.prefix_fmt)
            frame = _take_len(data, length)
            block = self.checksum.decode(frame)
            if block == self.EOF:
                log.debug('EOF frame detected')
                return

            yield block


def _take_fmt(data, fmt):
    length = struct.calcsize(fmt)
    chunk = bytearray(itertools.islice(data, length))
    if len(chunk) < length:
        raise ValueError('missing prefix data')
    return struct.unpack(fmt, bytes(chunk))


def _take_len(data, length):
    chunk = bytearray(itertools.islice(data, length))
    if len(chunk) < length:
        raise ValueError('missing payload data')
    return chunk


def chain_wrapper(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        return itertools.chain.from_iterable(result)
    return wrapped


class BitPacker(object):
    byte_size = 8

    def __init__(self):
        bits_list = []
        for index in range(2 ** self.byte_size):
            bits = [index & (2 ** k) for k in range(self.byte_size)]
            bits_list.append(tuple((1 if b else 0) for b in bits))

        self.to_bits = dict((i, bits) for i, bits in enumerate(bits_list))
        self.to_byte = dict((bits, i) for i, bits in enumerate(bits_list))


@chain_wrapper
def encode(data, framer=None):
    converter = BitPacker()
    framer = framer or Framer()
    for frame in framer.encode(data):
        for byte in frame:
            yield converter.to_bits[byte]


@chain_wrapper
def _to_bytes(bits):
    converter = BitPacker()
    for chunk in common.iterate(data=bits, size=8,
                                func=tuple, truncate=True):
        yield [converter.to_byte[chunk]]


@chain_wrapper
def decode(bits, framer=None):
    framer = framer or Framer()
    for frame in framer.decode(_to_bytes(bits)):
        yield frame



import numpy as np
import logging
import itertools
import collections

log = logging.getLogger(__name__)

from . import dsp
from . import equalizer
from . import common


class Detector(object):

    COHERENCE_THRESHOLD = 0.9

    CARRIER_DURATION = sum(equalizer.prefix)
    CARRIER_THRESHOLD = int(0.9 * CARRIER_DURATION)
    SEARCH_WINDOW = int(0.1 * CARRIER_DURATION)

    TIMEOUT = 10.0  # [seconds]

    def __init__(self, config, pylab):
        self.freq = config.Fc
        self.omega = 2 * np.pi * self.freq / config.Fs
        self.Nsym = config.Nsym
        self.Tsym = config.Tsym
        self.maxlen = config.baud  # 1 second of symbols
        self.max_offset = self.TIMEOUT * config.Fs
        self.plt = pylab

    def _wait(self, samples):
        counter = 0
        bufs = collections.deque([], maxlen=self.maxlen)
        for offset, buf in common.iterate(samples, self.Nsym, index=True):
            if offset > self.max_offset:
                raise ValueError('Timeout waiting for carrier')
            bufs.append(buf)

            coeff = dsp.coherence(buf, self.omega)
            if abs(coeff) > self.COHERENCE_THRESHOLD:
                counter += 1
            else:
                counter = 0

            if counter == self.CARRIER_THRESHOLD:
                return offset, bufs

        raise ValueError('No carrier detected')

    def run(self, samples):
        offset, bufs = self._wait(samples)

        length = (self.CARRIER_THRESHOLD - 1) * self.Nsym
        begin = offset - length

        x = np.concatenate(tuple(bufs)[-self.CARRIER_THRESHOLD:-1])
        Hc = dsp.exp_iwt(-self.omega, len(x))
        amplitude = np.abs(np.dot(Hc, x) / (0.5 * len(x)))
        start_time = begin * self.Tsym / self.Nsym
        log.info('Carrier detected at ~%.1f ms @ %.1f kHz:'
                 ' coherence=%.3f%%, amplitude=%.3f',
                 start_time * 1e3, self.freq / 1e3,
                 np.abs(dsp.coherence(x, self.omega)) * 100, amplitude)

        log.debug('Buffered %d ms of audio', len(bufs))

        bufs = list(bufs)[-self.CARRIER_THRESHOLD-self.SEARCH_WINDOW:]
        n = self.SEARCH_WINDOW + self.CARRIER_DURATION - self.CARRIER_THRESHOLD
        trailing = list(itertools.islice(samples, n * self.Nsym))
        bufs.append(np.array(trailing))

        buf = np.concatenate(bufs)
        offset = self.find_start(buf, duration=self.CARRIER_DURATION)
        start_time += (offset / self.Nsym - self.SEARCH_WINDOW) * self.Tsym
        log.debug('Carrier starts at %.3f ms', start_time * 1e3)

        buf = buf[offset:]

        prefix_length = self.CARRIER_DURATION * self.Nsym
        amplitude, freq_err = self.estimate(buf[:prefix_length])
        return itertools.chain(buf, samples), amplitude, freq_err

    def find_start(self, buf, duration):
        filt = dsp.FIR(dsp.exp_iwt(self.omega, self.Nsym))
        p = np.abs(list(filt(buf))) ** 2
        p = np.cumsum(p)[self.Nsym-1:]
        p = np.concatenate([[0], p])
        length = (duration - 1) * self.Nsym
        correlations = np.abs(p[length:] - p[:-length])
        offset = np.argmax(correlations)
        return offset

    def estimate(self, buf, skip=5):
        filt = dsp.exp_iwt(-self.omega, self.Nsym) / (0.5 * self.Nsym)
        frames = common.iterate(buf, self.Nsym)
        symbols = [np.dot(filt, frame) for frame in frames]
        symbols = np.array(symbols[skip:-skip])

        amplitude = np.mean(np.abs(symbols))
        log.debug('Carrier symbols amplitude : %.3f', amplitude)

        phase = np.unwrap(np.angle(symbols)) / (2 * np.pi)
        indices = np.arange(len(phase))
        a, b = dsp.linear_regression(indices, phase)
        self.plt.figure()
        self.plt.plot(indices, phase, ':')
        self.plt.plot(indices, a * indices + b)

        freq_err = a / (self.Tsym * self.freq)
        last_phase = a * indices[-1] + b
        log.debug('Current phase on carrier: %.3f', last_phase)
        log.debug('Frequency error: %.2f ppm', freq_err * 1e6)
        self.plt.title('Frequency drift: {0:.3f} ppm'.format(freq_err * 1e6))

        return amplitude, freq_err
